import os
import unittest
from .client import AuthsignalClient, ApiException


class TestAuthsignalClient(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_config = {
            "api_secret_key": os.getenv("AUTHSIGNAL_API_SECRET"),
            "api_url": os.getenv("AUTHSIGNAL_API_URL"),
        }

    def test_actions(self):
        client = AuthsignalClient(
            api_secret_key=self.test_config["api_secret_key"],
            api_url=self.test_config["api_url"],
        )

        track_attributes = {
            "redirect_url": "http://localhost:3000",
            "redirect_to_settings": True,
            "email": "not-a-real-email@authsignal.com",
            "phone_number": "1234567890",
            "ip_address": "127.0.0.1",
            "user_agent": "Authsignal-Python-SDK-Tests/1.0",
            "device_id": "device123",
            "custom": {"hello": "world"},
        }

        track_response = client.track(
            user_id="user123",
            action="python-sdk-test",
            attributes=track_attributes,
        )
        self.assertEqual(track_response["state"], "CHALLENGE_REQUIRED")
        self.assertTrue(track_response.get("idempotency_key"))
        self.assertTrue(track_response.get("is_enrolled", False))
        self.assertTrue(track_response.get("url"))
        self.assertTrue(track_response.get("token"))
        self.assertTrue(len(track_response.get("allowed_verification_methods", [])) > 0)

        get_action_response = client.get_action(
            user_id="user123",
            action="python-sdk-test",
            idempotency_key=track_response["idempotency_key"],
        )
        self.assertEqual(get_action_response["state"], "CHALLENGE_REQUIRED")
        self.assertIsNotNone(get_action_response.get("output"))

        update_action_response = client.update_action(
            user_id="user123",
            action="python-sdk-test",
            idempotency_key=track_response["idempotency_key"],
            attributes={"state": "BLOCK"},
        )
        self.assertEqual(update_action_response["state"], "BLOCK")

    def test_validate_challenge(self):
        client = AuthsignalClient(
            api_secret_key=self.test_config["api_secret_key"],
            api_url=self.test_config["api_url"],
        )

        track_response = client.track(
            user_id="user123",
            action="python-sdk-test",
            attributes={
                "redirect_url": "http://localhost:3000",
                "redirect_to_settings": True,
                "email": "not-a-real-email@authsignal.com",
                "phone_number": "1234567890",
                "ip_address": "127.0.0.1",
                "user_agent": "Authsignal-Python-SDK-Tests/1.0",
                "device_id": "device123",
            },
        )

        validate_response = client.validate_challenge(
            attributes={
                "token": track_response["token"],
            },
        )
        self.assertIsNotNone(validate_response.get("is_valid"))
        self.assertEqual(validate_response["state"], "CHALLENGE_REQUIRED")
        self.assertEqual(validate_response["user_id"], "user123")
        self.assertEqual(validate_response["action"], "python-sdk-test")
        self.assertTrue(validate_response.get("idempotency_key"))

    def test_authenticators(self):
        client = AuthsignalClient(
            api_secret_key=self.test_config["api_secret_key"],
            api_url=self.test_config["api_url"],
        )

        enroll_response = client.enroll_verified_authenticator(
            user_id="user12345",
            attributes={
                "verificationMethod": "EMAIL_OTP",
                "email": "not-a-real-email@authsignal.com",
            },
        )
        self.assertTrue(enroll_response["authenticator"]["user_authenticator_id"])
        self.assertEqual(
            enroll_response["authenticator"]["verification_method"],
            "EMAIL_OTP",
        )
        self.assertEqual(
            enroll_response["authenticator"]["email"],
            "not-a-real-email@authsignal.com",
        )
        self.assertEqual(enroll_response["authenticator"]["user_id"], "user12345")

        get_auth_response = client.get_authenticators(user_id="user12345")
        self.assertTrue(len(get_auth_response) > 0)

        authenticator = get_auth_response[0]
        self.assertTrue(authenticator["user_authenticator_id"])
        self.assertEqual(
            authenticator["verification_method"],
            "EMAIL_OTP",
        )
        self.assertEqual(authenticator["email"], "not-a-real-email@authsignal.com")

        client.delete_authenticator(
            user_id="user12345",
            user_authenticator_id=authenticator["user_authenticator_id"],
        )

        get_auth_response_after_delete = client.get_authenticators(user_id="user12345")
        for auth in get_auth_response_after_delete:
            self.assertNotEqual(
                auth["user_authenticator_id"], authenticator["user_authenticator_id"]
            )

    def test_new_client(self):
        client = AuthsignalClient("secret", "https://api.authsignal.com")
        self.assertEqual(client.api_secret_key, "secret")
        self.assertEqual(client.api_url, "https://api.authsignal.com")
        self.assertIsNotNone(client.session)

    def test_new_authsignal_api_error(self):
        error_code = "bad_request"
        error_description = "An error occurred"
        status_code = 400

        api_error = ApiException(
            error_code=error_code,
            error_description=error_description,
            status_code=status_code,
        )

        self.assertIsNotNone(api_error)
        self.assertEqual(api_error.error_code, error_code)
        self.assertEqual(api_error.error_description, error_description)
        self.assertEqual(api_error.status_code, status_code)

    def test_authsignal_api_error_message(self):
        status_code = 404
        error_description = "Not Found"

        api_error = ApiException(
            error_code="ERR404",
            error_description=error_description,
            status_code=status_code,
        )
        expected_error_message = "AuthsignalException: 404 - Not Found"
        self.assertEqual(str(api_error), expected_error_message)

    def test_users(self):
        client = AuthsignalClient(
            api_secret_key=self.test_config["api_secret_key"],
            api_url=self.test_config["api_url"],
        )

        update_user_response = client.update_user(
            user_id="a-new-user",
            attributes={
                "phoneNumber": "9876543210",
                "displayName": "A New User",
            },
        )

        self.assertEqual(
            update_user_response["phone_number"],
            "9876543210",
        )

        self.assertEqual(
            update_user_response["display_name"],
            "A New User",
        )

        get_user_response = client.get_user(user_id="a-new-user")
        self.assertEqual(
            get_user_response["phone_number"],
            "9876543210",
        )
        self.assertEqual(
            get_user_response["display_name"],
            "A New User",
        )

        client.delete_user(user_id="a-new-user")

        get_user_response = client.get_user(user_id="a-new-user")
        self.assertIsNotNone(get_user_response.get("is_enrolled"))
        self.assertFalse(get_user_response["is_enrolled"])

    def test_get_action_with_bad_secret(self):
        client = AuthsignalClient(
            api_secret_key="bad-secret",
            api_url=self.test_config["api_url"],
        )

        with self.assertRaises(ApiException) as cm:
            client.get_action(
                user_id="test-user",
                action="python-sdk-test",
                idempotency_key="test-key",
            )

        api_error = cm.exception
        self.assertEqual(api_error.status_code, 401)
        self.assertEqual(
            api_error.error_description,
            "The request is unauthorized. Check that your API key and region base URL are correctly configured.",
        )
        self.assertEqual(api_error.error_code, "unauthorized")
        self.assertEqual(
            str(api_error),
            "AuthsignalException: 401 - The request is unauthorized. Check that your API key and region base URL are correctly configured.",
        )

    def test_track_without_attributes(self):
        client = AuthsignalClient(
            api_secret_key=self.test_config["api_secret_key"],
            api_url=self.test_config["api_url"],
        )

        track_response = client.track(
            user_id="user123",
            action="python-sdk-test",
        )

        # Verify basic response structure
        self.assertEqual(track_response["state"], "CHALLENGE_REQUIRED")
        self.assertTrue(track_response.get("idempotency_key"))
        self.assertIsNotNone(track_response.get("token"))


if __name__ == "__main__":
    unittest.main()



import decimal
import json
import urllib.parse
from enum import Enum
from typing import Dict, Any

import humps
import requests
from requests.adapters import HTTPAdapter

from authsignal.version import VERSION

API_BASE_URL = "https://api.authsignal.com/v1"


class ActionState(Enum):
    BLOCK = "BLOCK"
    ALLOW = "ALLOW"
    CHALLENGE_REQUIRED = "CHALLENGE_REQUIRED"
    CHALLENGE_FAILED = "CHALLENGE_FAILED"
    CHALLENGE_SUCCEEDED = "CHALLENGE_SUCCEEDED"
    REVIEW_REQUIRED = "REVIEW_REQUIRED"
    REVIEW_FAILED = "REVIEW_FAILED"
    REVIEW_SUCCEEDED = "REVIEW_SUCCEEDED"


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return (str(o),)
        return super(DecimalEncoder, self).default(o)


class CustomSession(requests.Session):
    def __init__(self, timeout, api_key):
        super().__init__()
        self.mount("http://", HTTPAdapter())
        self.mount("https://", HTTPAdapter())

        self.timeout = timeout
        self.auth = requests.auth.HTTPBasicAuth(api_key, "")
        self.headers.update(
            {
                "Content-Type": "application/json",
                "Accept": "*/*",
                "User-Agent": "authsignal-python-sdk/" + VERSION,
                "X-Authsignal-Version": VERSION,
            }
        )

    def prepare_request(self, request):
        prepared_request = super().prepare_request(request)

        if (
            prepared_request.headers.get("Content-Type") == "application/json"
            and prepared_request.body
        ):
            data = json.loads(prepared_request.body)
            cleaned_data = self._remove_none_values(data)
            prepared_request.body = json.dumps(cleaned_data)
        return prepared_request

    @staticmethod
    def _remove_none_values(d: Dict[str, Any]) -> Dict[str, Any]:
        """Remove keys with None values from a dictionary."""
        return {k: v for k, v in d.items() if v is not None}

    def send(self, request, **kwargs) -> requests.Response:
        kwargs.setdefault("timeout", self.timeout)
        try:
            response = super().send(request, **kwargs)
            response.raise_for_status()

            if response.headers.get("Content-Type") == "application/json":
                data = response.json()
                decamelized_content = humps.decamelize(data)
                response.decamelized_content = decamelized_content
            return response
        except requests.exceptions.RequestException as e:
            error_code = None
            error_description = None
            status_code = None

            if isinstance(e, requests.exceptions.HTTPError):
                status_code = e.response.status_code
                try:
                    error_data = e.response.json()
                    error_code = error_data.get("errorCode")
                    error_description = error_data.get("errorDescription")
                except (ValueError, AttributeError):
                    pass

            raise ApiException(error_code, error_description, status_code) from e


class AuthsignalClient(object):

    def __init__(self, api_secret_key, api_url=API_BASE_URL, timeout=2.0):
        """Initialize the client.
        Args:
            api_key: Your Authsignal Secret API key of your tenant
            api_url: Base URL, including scheme and host, for sending events.
                Defaults to 'https://api.authsignal.com/v1'.
            timeout: Number of seconds to wait before failing request. Defaults
                to 2 seconds.
        """
        _assert_non_empty_string(api_url, "api_url")
        _assert_non_empty_string(api_secret_key, "api_secret_key")

        self.api_secret_key = api_secret_key
        self.api_url = api_url

        self.session = CustomSession(timeout=timeout, api_key=api_secret_key)
        self.version = VERSION

    def track(
        self, user_id: str, action: str, attributes: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        """Tracks an action to authsignal, scoped to the user_id and action
        Returns the status of the action so that you can determine to whether to continue
        Args:
            user_id:  A user's id. This id should be the same as the user_id used in
                event calls.
            action: The action that you are tracking an event for, i.e. signIn.
            attributes: A dictionary containing the request body. Optional.
        """
        _assert_non_empty_string(user_id, "user_id")
        _assert_non_empty_string(action, "action")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}/actions/{urllib.parse.quote(action)}"

        attributes = attributes or {}
        response = self.session.post(
            url=path, data=json.dumps(attributes, cls=DecimalEncoder)
        )

        return response.decamelized_content

    def get_user(self, user_id: str) -> Dict[str, Any]:
        """Retrieves the user from authsignal
        Args:
            user_id:  A user's id.
        """
        _assert_non_empty_string(user_id, "user_id")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}"

        response = self.session.get(url=path)

        return response.decamelized_content

    def update_user(self, user_id: str, attributes: Dict[str, Any]) -> Dict[str, Any]:
        """Updates the user in authsignal
        Args:
            user_id:  A user's id.
            attributes: A dictionary containing the request body.
        """
        _assert_non_empty_string(user_id, "user_id")
        _assert_non_empty_dict(attributes, "attributes")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}"

        response = self.session.patch(
            url=path, data=json.dumps(attributes, cls=DecimalEncoder)
        )

        return response.decamelized_content

    def delete_user(self, user_id: str):
        """Deletes a user from authsignal
        Args:
            user_id:  A user's id.
        """
        _assert_non_empty_string(user_id, "user_id")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}"

        response = self.session.delete(url=path)

        return

    def get_authenticators(self, user_id: str) -> Dict[str, Any]:
        """Retrieves the authenticators for a user
        Args:
            user_id:  A user's id.
        """
        _assert_non_empty_string(user_id, "user_id")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}/authenticators"

        response = self.session.get(url=path)

        return response.decamelized_content

    def enroll_verified_authenticator(
        self, user_id: str, attributes: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Enrolls an authenticator for a given user.
        Args:
            user_id:  A user's id. This id should be the same as the user_id used in event calls.
            attributes:  A dictionary containing the request body.
        """
        _assert_non_empty_string(user_id, "user_id")
        _assert_non_empty_dict(attributes, "attributes")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}/authenticators"

        response = self.session.post(
            url=path, data=json.dumps(attributes, cls=DecimalEncoder)
        )

        return response.decamelized_content

    def delete_authenticator(self, user_id: str, user_authenticator_id: str):
        """Deletes an authenticator from authsignal
        Args:
            user_id: A user's id.
            user_authenticator_id: The id of the authenticator you want to delete
        """
        _assert_non_empty_string(user_id, "user_id")
        _assert_non_empty_string(user_authenticator_id, "user_authenticator_id")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}/authenticators/{urllib.parse.quote(user_authenticator_id)}"

        response = self.session.delete(url=path)

        return

    def validate_challenge(self, attributes: Dict[str, Any]) -> Dict[str, Any]:
        """Validates a token from authsignal
        Args:
            attributes: A dictionary containing the token to validate.
        """
        _assert_non_empty_dict(attributes, "attributes")

        path = f"{self.api_url}/validate"

        response = self.session.post(
            url=path, data=json.dumps(attributes, cls=DecimalEncoder)
        )

        return response.decamelized_content

    def get_action(
        self, user_id: str, action: str, idempotency_key: str
    ) -> Dict[str, Any]:
        """Retrieves the action from authsignal for a given user and action.
        Args:
            user_id: A user's id.
            action: The action that you are retrieving, i.e. signIn
            idempotency_key: The action's idempotency key
        """
        _assert_non_empty_string(user_id, "user_id")
        _assert_non_empty_string(action, "action")
        _assert_non_empty_string(idempotency_key, "idempotency_key")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}/actions/{urllib.parse.quote(action)}/{urllib.parse.quote(idempotency_key)}"

        response = self.session.get(url=path)

        return response.decamelized_content

    def update_action(
        self,
        user_id: str,
        action: str,
        idempotency_key: str,
        attributes: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Updates an action in authsignal
        Args:
            user_id: A user's id.
            action: The action that you are updating, i.e. signIn
            idempotency_key: The action's idempotency key
            attributes: A dictionary containing the request body.
        """
        _assert_non_empty_string(user_id, "user_id")
        _assert_non_empty_string(action, "action")
        _assert_non_empty_string(idempotency_key, "idempotency_key")
        _assert_non_empty_dict(attributes, "attributes")

        path = f"{self.api_url}/users/{urllib.parse.quote(user_id)}/actions/{urllib.parse.quote(action)}/{urllib.parse.quote(idempotency_key)}"

        response = self.session.patch(
            url=path, data=json.dumps(attributes, cls=DecimalEncoder)
        )

        return response.decamelized_content


class ApiException(Exception):
    def __init__(self, error_code, error_description, status_code):
        super().__init__(f"AuthsignalException: {status_code} - {error_description}")
        self.error_code = error_code
        self.error_description = error_description
        self.status_code = status_code

    def __str__(self):
        return f"AuthsignalException: {self.status_code} - {self.error_description}"


def _assert_non_empty_string(val: str, name: str) -> None:
    if not isinstance(val, str) or not val:
        raise ValueError(f"{name} must be a non-empty string")


def _assert_non_empty_dict(val: dict, name: str) -> None:
    if not isinstance(val, dict) or not val:
        raise ValueError(f"{name} must be a non-empty dict")
